REBOL [
	Name: 		"Printer Scheme"
	Purpose:	"Printing support library for Windows"
	Author:		"Nenad Rakocevic / SOFTINNOV"
	Email:		nr@softinnov.com
	File: 		%printer-scheme.r
	Version:	0.9.0
	Date:		08/09/2008
	Needs: 		[library]
	License:	"BSD"
]

make root-protocol [
	scheme: 'printer

	sw: system/words
	
	open*: 	 	get in sw 'open
	copy*: 	 	get in sw 'copy
	insert*: 	get in sw 'insert
	close*:  	get in sw 'close
	get-modes*: get in sw 'get-modes
	query*:		get in sw 'query
	
	net-log: 	get in net-utils 'net-log

	open-proto: none
	port-flags: system/standard/port-flags/pass-thru
	
	set 'printer context [
		defs: [
			caps [
				TECHNOLOGY		2
				HORZSIZE		4
				VERTSIZE		6
				HORZRES			8
				VERTRES			10
				LOGPIXELSX		88
				LOGPIXELSY	 	90
				BITSPIXEL 		12
				PDEVICESIZE 	26
				CLIPCAPS 		36
				RASTERCAPS		38
				SIZEPALETTE 	104
				NUMRESERVED 	106
				COLORRES 		108
				PHYSICALWIDTH 	110
				PHYSICALHEIGHT 	111
				PHYSICALOFFSETX 112
				PHYSICALOFFSETY 113
				SCALINGFACTORX 	114
				SCALINGFACTORY 	115
				VREFRESH 		116
				DESKTOPHORZRES	118
				DESKTOPVERTRES	117
				BLTALIGNMENT	119
			]
			dt [
				LEFT 			0
				TOP  			0
				CENTER       	1
				RIGHT        	2
				VCENTER      	4
				BOTTOM       	8
				WORDBREAK    	16
				SINGLELINE   	32
				EXPANDTABS  	64
				TABSTOP      	128
				NOCLIP       	256
				EXTERNALLEADING 512
				CALCRECT     	1024
				NOPREFIX     	2048
				INTERNAL     	4096
				EDITCONTROL  	8192
				PATH_ELLIPSIS   16384
				END_ELLIPSIS 	32768
				MODIFYSTRING 	65536
				RTLREADING   	131072
				WORD_ELLIPSIS 	262144
			]
			pen [
				solid	  		0
				dash			1
				dot				2
				dashdot			3
				dashdotdot		4
				null			5
				insideframe		6
				userstyle		7
				alternate		8

				endcap_round	0
				endcap_square	256
				endcap_flat		512

				join-round		0
				join-bevel		4096
				join-miter		8192
			]
			stock [
				white-brush 	0
				lt-gray-brush 	1
				gray-brush 		2
				dk-gray-brush 	3
				black-brush 	4
				null-brush 		5
				hollow-brush 	5
				white-pen		6
				black-pen		7
				null-pen		8
				system-font		13
				default-palette 15
			]
			printer-enum [
				local 	  		2 
				connections 	4
				remote	  		16
			]
			misc [
				DIB_RGB_COLORS	0
				SRCCOPY			13369376 
				DM_OUT_BUFFER 	2
			]
		]

		kernel32:	load/library %kernel32.dll
		gdi32: 		load/library %gdi32.dll
		user32: 	load/library %user32.dll
		winspool: 	load/library %winspool.drv

		; === General API ===

		int: :to-integer

		FORMAT_MESSAGE_FROM_SYSTEM:	   int #{00001000}
		FORMAT_MESSAGE_IGNORE_INSERTS: int #{00000200}
		fmt-msg-flags: 
			FORMAT_MESSAGE_FROM_SYSTEM
			or FORMAT_MESSAGE_IGNORE_INSERTS

		GetLastError: make routine! [
			return: [integer!]
		] kernel32 "GetLastError"

		FormatMessage: make routine! [
			dwFlags		 [integer!]
			lpSource	 [integer!]
			dwMessageId  [integer!]
			dwLanguageId [integer!]
			lpBuffer	 [string!]
			nSize		 [integer!]
			Arguments	 [integer!]
			return:		 [integer!]
		] kernel32 "FormatMessageA"

		; === Printer API ===

		_RECT: make struct! struct-rect: [
			left		[integer!]
			top			[integer!]
			right		[integer!]
			bottom		[integer!]
		] none

		DOC_INFO: make struct! struct-doc-info: [
			[save]
			cbSize			[integer!]
			lpszDocName		[string!]
			lpszOutput		[integer!]
			lpszDatatype 	[integer!]
			fwType			[integer!]
		][0 "test" 0 0 0]

		PRINTER_INFO_4: make struct! [
			pPrinterName	[string!]
			pServerName		[string!]
			Attributes		[integer!]
		] none

		TEXTMETRIC: make struct! struct-text-metric: [
			tmHeight			[long]
			tmAscent			[long]
			tmDescent			[long]
			tmInternalLeading	[long]
			tmExternalLeading	[long]
			tmAveCharWidth		[long]
			tmMaxCharWidth		[long]
			tmWeight			[long]
			tmOverhang			[long]
			tmDigitizedAspectX	[long]
			tmDigitizedAspectY	[long] 
			tmFirstChar			[char]
			tmLastChar			[char]
			tmDefaultChar		[char]
			tmBreakChar			[char]
			tmItalic			[char]
			tmUnderlined		[char]
			tmStruckOut			[char]
			tmPitchAndFamily	[char]
			tmCharSet			[char]
		] none

		BITMAPINFOHEADER: make struct! bih: [
			biSize			[long]
			biWidth			[long]
			biHeight		[long]
			biPlanes		[short]
			biBitCount		[short]
			biCompression	[long] 
			biSizeImage		[long] 
			biXPelsPerMeter	[long] 
			biYPelsPerMeter	[long] 
			biClrUsed		[long] 
			biClrImportant	[long] 
		] none

		BITMAPINFO: make struct! struct-bitmap-info: append copy* bih [
			bmiColors		[integer!]
		] none

		GetDefaultPrinter: make routine! [
			pszBuffer	[string!]
			pcchBuffer  [struct! [low [integer!] high [integer!]]]
			return: 	[integer!]
		] winspool "GetDefaultPrinterA"

		OpenPrinter: make routine! [
			pPrinterName [string!]
			phPrinter  	[struct! [ptr [integer!]]]
			pDefault  	[integer!]
			return: 	[integer!]
		] winspool "OpenPrinterA"

		EnumPrinters: make routine! [
			Flags 		[integer!]
			Name 		[integer!]
			Level 		[integer!]
			pPrinterEnum [string!]
			cbBuf 		[integer!]
			pcbNeeded	[struct! [n [integer!]]]
			pcReturned  [struct! [n [integer!]]]
			return: 	[integer!]
		] winspool "EnumPrintersA"

		ClosePrinter: make routine! compose/deep [
			hPrinter	[integer!]
			return: 	[integer!]
		] winspool "ClosePrinter"

		StartDoc: make routine! compose/deep [
			hdc			[integer!]
			lpdi		[struct! [(struct-doc-info)]]   
			return: 	[integer!]
		] gdi32 "StartDocA"

		StartPage: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "StartPage"

		EndPage: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "EndPage"

		EndDoc: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "EndDoc"

		CreateDC: make routine! [
			lpszDriver	[string!]
			lpszDevice	[string!]
			lpszOutput	[integer!]
			lpInitData	[integer!]
			return: 	[integer!]
		] gdi32 "CreateDCA"

		DeleteDC: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "DeleteDC"

		TextOut: make routine! [
		  hdc			[integer!]
		  nXStart		[integer!]
		  nYStart		[integer!]
		  lpString		[string!]
		  cbString		[integer!]
		  return: 		[integer!]
		] gdi32 "TextOutA"

		DrawText: make routine! compose/deep [
			hdc			[integer!]
			lpString	[string!]
			nCount		[integer!]
			lpRect		[struct!  [(struct-rect)]] 
			uFormat		[integer!]
			return: 	[integer!]
		] user32 "DrawTextA"

		SetTextColor: make routine! [
			hdc			[integer!]
			crColor		[integer!]
			return: 	[integer!]
		] gdi32 "SetTextColor"

		MoveToEx: make routine! [
			hdc			[integer!]
			X			[integer!]
			Y			[integer!]
			lpPoint		[integer!]
			return: 	[integer!]
		] gdi32 "MoveToEx"

		LineTo: make routine! [
			hdc			[integer!]
			nXEnd		[integer!]
			nYEnd		[integer!]
			return: 	[integer!]
		] gdi32 "LineTo"

		Arc: make routine! [
			hdc			[integer!]
			nLeftRect	[integer!]
			nTopRect	[integer!]
			nRightRect	[integer!]
			nBottomRect	[integer!]
			nXRadial1	[integer!]
			nYRadial1	[integer!]
			nXRadial2	[integer!]
			nYRadial2	[integer!]
			return: 	[integer!]
		] gdi32 "Arc" 

		Pie: make routine! third :Arc gdi32 "Pie" 

		Rectangle: make routine! [
			hdc			[integer!]
			nLeftRect	[integer!]
			nTopRect	[integer!]
			nRightRect	[integer!]
			nBottomRect	[integer!]
			return: 	[integer!]
		] gdi32 "Rectangle"

		CreatePen: make routine! [
			fnPenStyle	[integer!]
			nWidth		[integer!]
			crColor		[integer!]
			return: 	[integer!]
		] gdi32 "CreatePen"

		CreateSolidBrush: make routine! [
			crColor 	[integer!]
			return: 	[integer!]
		] gdi32 "CreateSolidBrush"

		GetDCBrushColor: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "GetDCBrushColor"

		GetDCPenColor: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "GetDCPenColor"

		GetTextColor: make routine! [
			hdc			[integer!]
			return: 	[integer!]
		] gdi32 "GetTextColor"

		GetStockObject: make routine! [
			fnObject	[integer!]
			return: 	[integer!]
		] gdi32 "GetStockObject"

		SelectObject: make routine! [
			hdc			[integer!]
			hgdiobj		[integer!]
			return: 	[integer!]
		] gdi32 "SelectObject"

		DeleteObject: make routine! [
			hgdiobj		[integer!]
			return: 	[integer!]
		] gdi32 "DeleteObject"

		StretchDIBits: make routine! compose/deep [
			hdc			[integer!]
			XDest		[integer!]
			YDest		[integer!]
			nDestWidth	[integer!]
			nDestHeight	[integer!]
			XSrc		[integer!]
			YSrc		[integer!]
			nSrcWidth	[integer!]
			nSrcHeight	[integer!]
			lpBits		[string!]
			lpBitsInfo	[struct! [(struct-bitmap-info)]]
			iUsage		[integer!]
			dwRop		[long]
			return: 	[integer!]
		] gdi32 "StretchDIBits"

		CreateFont: make routine! [
			nHeight				[integer!]
			nWidth				[integer!]
			nEscapement			[integer!]
			nOrientation		[integer!]
			fnWeight			[integer!]
			fdwItalic			[integer!]
			fdwUnderline		[integer!]
			fdwStrikeOut		[integer!]
			fdwCharSet			[integer!]
			fdwOutputPrecision	[integer!]
			fdwClipPrecision	[integer!]
			fdwQuality			[integer!]
			fdwPitchAndFamily	[integer!]
			lpszFace			[string!]
			return: 			[integer!]
		] gdi32 "CreateFontA"

		GetTextMetrics: make routine! compose/deep [
			hdc			[integer!]
			lptm		[struct! [(struct-text-metric)]]
			return: 	[integer!]
		] gdi32 "GetTextMetricsA"

		GetDeviceCaps: make routine! [
			hdc		[integer!]
			nIndex	[integer!]
			return: [integer!]
		] gdi32 "GetDeviceCaps"

		; === Helper functions ====
		
		locals: none
				
		rect: make struct! _RECT none
	
		make-null-string!: func [len [integer!]][
			head insert*/dup make string! len null len
		]

		get-error-msg: has [out][
			out: make-null-string! 256
			FormatMessage fmt-msg-flags 0 GetLastError 0 out 256 0
			trim/tail out
		]

		try*: func [body [block!] /quiet /local res][
			if all [zero? res: do body not quiet][
				print reduce [mold first body "failed :" get-error-msg]
			]
			res
		]

		to-xdpi: func [v [integer!]][
			to integer! v * locals/caps/LOGPIXELSX / 25.4 * locals/scale/x / 100
		]
		to-ydpi: func [v [integer!]][
			to integer! v * locals/caps/LOGPIXELSY / 25.4 * locals/scale/y / 100
		]

		to-pt-dpi: func [v][
			negate to integer! v * locals/caps/LOGPIXELSY / 72 * locals/scale/y / 100
		]

		to-bgr: func [c [tuple! word!]][
			if word? c [c: get c]	
			(c/3 * 65536) + (c/2 * 256) + c/1
		]

		to-tuple: func [bgr [integer!]][
			to tuple! reduce [
				255 and bgr
				255 and shift bgr 8
				255 and shift bgr 16
			]
		]

		update-obj: func [type [word!] /local obj][
			obj: locals/:type			
			if obj/dirty? [
				if obj/handle [try* [DeleteObject obj/handle]]
				obj/handle: switch type [
					pen	  [CreatePen obj/style to-pt-dpi obj/width to-bgr obj/color]
					brush [CreateSolidBrush to-bgr obj/color]
				]
				SelectObject locals/hDC obj/handle
				obj/dirty?: no
			]
		]

		reset-obj: func [type [word!] /local obj][
			obj: locals/:type
			if obj/handle [		
				SelectObject locals/hDC GetStockObject select defs/stock pick [null-pen null-brush] type = 'pen
				try* [DeleteObject obj/handle]
			]
			obj/handle: none
			obj/dirty?: no
		]

		enum: has [out buf flags needed ret cmd pi4 len buf*][
			out: 	make block! 4
			buf: 	make-null-string! 1024
			needed: make struct! [n [integer!]] none
			ret: 	make struct! needed none
			flags: 	defs/printer-enum/local or defs/printer-enum/connections

			try* cmd: [EnumPrinters flags 0 4 buf 1024 needed ret]

			pi4: make struct! PRINTER_INFO_4 none
			len: length? buf*: third pi4
			if 1024 < (ret/n * len) [
				buf: make string! ret/n * len
				try* cmd
			]
			loop ret/n [
				change/part buf* buf at buf len
				append/only out ret: second pi4
				poke ret 3 pick [remote local] zero? ret/3 and 64
				buf: at buf len + 1
			]
			buf: none
			new-line/all out on
		]

		init: func [/with prn-name /local len buf size][
			either prn-name [
				locals/name: prn-name
			][
				size: make struct! [low [integer!] high [integer!]] reduce [128 0]
				try* [GetDefaultPrinter locals/name size]
				clear find locals/name null
			]		
			locals/hDC: try* [CreateDC "WINSPOOL" locals/name 0 0]
			; caps => object with rebolish names
			locals/caps: reduce [
				'LOGPIXELSX 		GetDeviceCaps locals/hDC defs/caps/LOGPIXELSX
				'LOGPIXELSY 		GetDeviceCaps locals/hDC defs/caps/LOGPIXELSY
				'PHYSICALWIDTH		GetDeviceCaps locals/hDC defs/caps/PHYSICALWIDTH
				'PHYSICALHEIGHT		GetDeviceCaps locals/hDC defs/caps/PHYSICALHEIGHT
				'PHYSICALOFFSETX	GetDeviceCaps locals/hDC defs/caps/PHYSICALOFFSETX
				'PHYSICALOFFSETY	GetDeviceCaps locals/hDC defs/caps/PHYSICALOFFSETY
				'HORZRES			GetDeviceCaps locals/hDC defs/caps/HORZRES
				'VERTRES			GetDeviceCaps locals/hDC defs/caps/VERTRES
			]
			if locals/auto-fit?/x = 1 [
				locals/scale/x: to integer! 100 * locals/caps/HORZRES / locals/caps/PHYSICALWIDTH
			]
			if locals/auto-fit?/y = 1 [
				locals/scale/y: to integer! 100 * locals/caps/VERTRES / locals/caps/PHYSICALHEIGHT
			]
			locals/pen/color:   to-tuple GetDCPenColor locals/hDC
			locals/brush/color: to-tuple GetDCBrushColor locals/hDC
			SelectObject locals/hDC GetStockObject defs/stock/null-brush
		]

		draw-image: func [
			img  [image! word!]
			pos  [pair! none!]
			size [pair! none!]
			/local
				data out len pad pad-sz
		][
			if word? img [img: get img]

			;-- RGB->GRB conversion
			data: img
			forall data [change/only data reverse/part data/1 2 2]
			data: img/rgb

			;-- padding each scanline (4 bytes aligned)
			len: 3 * img/size/x
			pad: all [
				not zero? pad-sz: len // 4
				head insert*/dup copy* #{} null 4 - pad-sz
			]
			out: make string! img/size/y * pad-sz + length? data 
			repeat y img/size/y [
				insert*/part tail out at data (y - 1) * len len
				if pad [insert* tail out pad]
			]

			BITMAPINFO/biSize: length? third BITMAPINFOHEADER
			BITMAPINFO/biWidth: img/size/x
			BITMAPINFO/biHeight: negate img/size/y
			BITMAPINFO/biPlanes: 1
			BITMAPINFO/biBitCount: 24
	; Todo: convert img/size (px) to log.units
			try* [
				StretchDIBits
					locals/hDC
					to-xdpi pos/x
					to-ydpi	pos/y
					to-xdpi to integer! img/size/x * locals/scale/x / 400 ; hacked conversion
					to-ydpi to integer! img/size/y * locals/scale/y / 400 ; hacked conversion
					0 0 img/size/x img/size/y
					out
					BITMAPINFO
					defs/misc/DIB_RGB_COLORS
					defs/misc/SRCCOPY 
			]
		]

		draw-text: func [txt [string!] pos [pair!]][
			try* [TextOut locals/hDC to-xdpi pos/1 to-ydpi pos/2 txt length? txt]
		]

		draw-text-box: func [txt [string!] pos [pair!] sz [pair!] opts [block! none!] /local v][
			rect/left:	 to-xdpi pos/x
			rect/top:	 to-ydpi pos/y
			rect/right:  to-xdpi pos/x + sz/x
			rect/bottom: to-ydpi pos/y + sz/y				
			v: defs/dt/NOPREFIX or defs/dt/EXPANDTABS
			if not all [opts find opts 'no-wrap][v: v or defs/dt/WORDBREAK]
			v: v or any [select defs/dt locals/cur-font/align 0]
			v: v or any [
				either locals/cur-font/valign = 'center [defs/dt/VCENTER][
					select defs/dt locals/cur-font/valign
				]
				0
			]
			try* [DrawText locals/hDC txt -1 rect v]
		]

		draw-line: func [pos [pair!] pos2 [pair!]][
			update-obj 'pen
			try* [MoveToEx locals/hDC to-xdpi pos/x  to-ydpi pos/y 0]
			try* [LineTo   locals/hDC to-xdpi pos2/x to-ydpi pos2/y]
		]

		draw-box: func [pos [pair!] pos2 [pair!]][
			update-obj 'pen
			update-obj 'brush			
			try* [Rectangle locals/hDC to-xdpi pos/x to-ydpi pos/y to-xdpi pos2/x to-ydpi pos2/y]
		]

		draw-arc: func [center [pair!] radius [pair!] beg [decimal!] len [decimal!] closed [logic!] /local cmd][
			update-obj 'pen			
			update-obj 'brush
			;beg: beg - 180
			cmd: [
				? 
				locals/hDC
				to-xdpi center/x - radius/x
				to-ydpi center/y - radius/y
				to-xdpi center/x + radius/x
				to-ydpi center/y + radius/y
				to-xdpi to integer! center/x + (radius/x * cosine beg)
				to-ydpi to integer! center/y - (radius/y * sine beg)
				to-xdpi to integer! center/x + (radius/x * cosine beg + len)
				to-ydpi to integer! center/y - (radius/y * sine beg + len)
			]
			poke cmd 1 pick [Pie Arc] to logic! closed
			try* cmd
		]

		set-font: func [obj [word! object!] /local v fnt][
			if any-word? obj [obj: get obj]
			if not fnt: select locals/font-cache obj [
				v: obj/style
				fnt: try* [
					CreateFont
						to-pt-dpi obj/size
						0 0 0
						either any [v = 'bold all [series? v find v 'bold]][700][400]
						either any [v = 'italic all [series? v find v 'italic]][-1][0]
						either any [v = 'underline all [series? v find v 'underline]][-1][0]
						0 0	0 0	0 0
						obj/name
				]	
				repend locals/font-cache [obj fnt]
			]
			try* [SelectObject locals/hDC fnt]
			if locals/cur-font/color <> obj/color [SetTextColor locals/hDC to-bgr obj/color]
			locals/cur-font: obj
		]

		set-pen: func [c1 [tuple! word!] c2 [tuple! none!]][
			locals/pen/color: c1
			if c2 [locals/pen/color2: c2]
			locals/pen/dirty?: yes
		]

		set-fill: func [c1 [tuple! word!]][
			locals/brush/color: c1
			locals/brush/dirty?: yes
		]

		set-line-width: func [size [integer!]][
			locals/pen/width: size
			locals/pen/dirty?: yes
		]

		get-font-metrics: has [fnt][
			try* [GetTextMetrics locals/hDC TEXTMETRIC]
			TEXTMETRIC
		]

		get-drawable-size: func [/mm /local caps][
			caps: locals/caps
			as-pair 
				either mm [caps/HORZRES * 254 / caps/LOGPIXELSX / 10][caps/HORZRES]
				either mm [caps/VERTRES * 254 / caps/LOGPIXELSY / 10][caps/VERTRES]
		]

		start-doc:	func [/title name][
			if title [DOC_INFO/lpszDocName: name]
			DOC_INFO/cbSize: length? third DOC_INFO
			try* [StartDoc locals/hDC DOC_INFO]
		]
		start-page: does [try* [StartPage locals/hDC]]
		end-page:	does [try* [EndPage locals/hDC]]
		end-doc:	does [try* [EndDoc locals/hDC]]

		emit: func [spec [block!] /local pos pos2 txt obj sz opts c1 c2 rad][
			locals/cur-font: face/font		
			parse spec [
				some [
					'text 
						set txt		string!
						set pos		pair! 
						(draw-text txt pos)

					| 'text-box 
						set txt		string!
						set pos		pair!
						set sz		pair!
						opt [set opts block!]
						(draw-text-box txt pos sz opts opts: none)

					| 'line 
						set pos 	pair!
						set pos2 	pair! 
						(draw-line pos pos2)

					| 'box 
						set pos 	pair!
						set pos2 	pair! 
						(draw-box pos pos2)

					| 'arc
						set	pos 	pair!
						set rad 	pair!
						set c1		[decimal! | integer!]
						set c2 		[decimal! | integer!]
						['closed (opts: yes) | none]
						(draw-arc pos rad to decimal! c1 to decimal! c2 to-logic opts opts: none)

					| 'font 
						set obj [word! | object!] 
						(set-font obj)

					| 'pen 
						set c1 [tuple! | word!]
						[set c2 tuple! | none]
						(set-pen c1 c2 c2: none)

					| 'fill-pen
						set c1 [tuple! | word!]
						(either c1 = 'none [reset-obj 'brush][set-fill c1])

					| 'line-width
						set sz integer!
						(set-line-width sz)

					| 'image 
						set obj [image! | word!]
						opt [set pos  [pair!]]
						opt [set pos2 [pair!]]
						(draw-image obj pos pos2 pos: pos2: none)

					| skip
				]
			]
		]

		emit-page: func [spec [block!]][
			start-page
			emit spec
			end-page
			;forall spec [if pair? spec/1 [change spec spec/1 * 2.5]]
			;spec: head spec
			;view/new layout compose/deep [
			;	size 800x1000
			;	origin 0x0
			;	backcolor white
			;	box 800x1000 effect [draw [(spec)]]
			;]
		]

		emit-doc: func [spec [block!]][
			start-doc
			emit-page spec
			end-doc
		]

		close: does [
			SelectObject locals/hDC 0	;-- release last selected handle
			foreach [obj hfnt] locals/font-cache [try* [DeleteObject hfnt]]
			if locals/pen/handle   [try* [DeleteObject locals/pen/handle]]
			if locals/brush/handle [try* [DeleteObject locals/brush/handle]]
			clear locals/font-cache
			try* [DeleteDC locals/hDC]
		]

	]
	
	;------- External API -------

	init: func [port spec][		
		if url? spec [net-utils/url-parser/parse-url port spec]
		port/url: spec
		port/locals: context [
			hDC: none
			hPrinter: make struct! [ptr [integer!]] none
			name: printer/make-null-string! 128
			font-cache: make hash! 8
			cur-font: face/font
			pen: copy* [
				handle  #[none]
				style	0
				width	1
				color	0.0.0
				color2	0.0.0
				dirty?	#[false]
			]
			brush: copy* [
				handle  #[none]
				color	255.255.255
				dirty?	#[false]
			]
			caps: none					; printer metrics
			scale: 100x100				; scaling X and Y factors
			auto-fit?: 1x1				; auto-scaling in X and Y to fit in drawable area (0x0 disables it)
			session: [
				started? [#[false] #[false]]
				pages 	0
				doc-name "no-name"
			]
			pages: make block! 1
		]
		port
	]

	open: func [port /local svc psc name][
		printer/locals: port/locals
		port/state/flags: port/state/flags or port-flags

		psc: port/state/custom
		either all [psc	name: select port/state/custom 'printer][
			printer/init/with name
		][
			printer/init
		]
		if all [psc name: select psc 'doc-name][
			port/locals/session/doc-name: name
		]
		;port/state/tail: 1 ;-- hack to enable 'remove calls
		port
	]
	
	insert: func  [port spec [word! block!] /local pl direct?][
		pl: printer/locals: port/locals
		if any [
			spec = 'start-doc
			not pl/session/started?/1 
		][
			direct?: yes
			printer/start-doc/title pl/session/doc-name
			pl/session/started?/1: yes
		]
		if any [
			spec = 'start-page
			not pl/session/started?/2 
		][		
			printer/start-page
			pl/session/started?/2: yes
		]
		if block? spec [
			printer/emit spec
		]
		if any [direct?	spec = 'end-doc][
			printer/end-doc
			pl/session/started?/1: no
		]
		if any [direct? spec = 'end-doc][
			printer/end-doc
			pl/session/started?/1: no
		]
		port
	]
	
	query: copy: func [port][make error! "unsupported function!"]
		
	close: func [port /local pl][
		printer/locals: port/locals
		printer/close
		port/state/flags: 0
	]
	
	get-modes: func [port spec /local direct?][
		printer/locals: port/locals
		;-- To be done
		if direct? [close port]
	]
	
	set-modes: func [port spec /local swc more? direct?][
		printer/locals: port/locals
		;-- To be done
		if direct? [close port]
	]
		
	net-utils/net-install printer self none
]

text-box: 'text 	;-- hack for backward compatibility with Draw dialect

protect 'printer	;-- printer's low-level functions are exposed through 'printer object
					;-- in case someone want to create a different input dialect